"""Fixtures definitions for pyEGPS tests."""
from __future__ import annotations

from array import array
from unittest.mock import MagicMock

import pytest

CTRL_IN = 1 << 7


def get_usb_mock(product_id: int) -> MagicMock:
    """Create a Mock representing a pyusb device."""
    mock = MagicMock()
    mock.idProduct = product_id
    mock.manufacturer = "AllFake"
    mock.product = "Fake-Product"

    def ctrl_transfer(
        bmRequestType,
        bRequest,
        wValue,
        wIndex,
        data_or_wLength,
        USB_CTRL_TRANSFER_TIMEOUT,
    ) -> bytes | int:
        req_in: bool = (bmRequestType & CTRL_IN) == CTRL_IN
        return array("B", []) if req_in else 0

    mock.ctrl_transfer = ctrl_transfer
    return mock


@pytest.fixture
def fakeUsbDevice() -> MagicMock:
    """Return a fake USB device with supported product_id."""
    return get_usb_mock(product_id=0xFD15)


@pytest.fixture
def fakeUnknownUsbDevice() -> MagicMock:
    """Return a fake USB device with unsupported product_id."""
    return get_usb_mock(product_id=0xFD20)



"""Allows direct execution of CLI with dummy devices."""
from unittest.mock import patch
import sys
import os

sys.path.append( os.path.join(os.path.dirname(__file__), "..","src") )

import pyegps
from pyegps.fakes.powerstrip import FakePowerStrip

def take_dummy_devices():
    return patch("pyegps.search_for_devices", return_value=FakePowerStrip.search_for_devices() )


if __name__ == "__main__":
    with take_dummy_devices() as mock:
        from pyegps import cli
        sys.exit(cli.cli(sys.argv[1:]))



"""Definition of the base class for alle devices."""
from __future__ import annotations

import abc

from typing import TypeVar

T = TypeVar("T", bound="Device")


class Device(abc.ABC):
    """Abstract base class for all devices."""

    def release(self) -> None:
        """Free any resources if needed."""

    @property
    def device_id(self) -> str:
        """Return a unique identifier among all devices of any type."""
        return f"{self.get_implementation_id()}:{self.uid}"

    @classmethod
    def is_responsible(cls, device_id: str) -> bool:
        """Check if the given device_id is handled by this device implementation."""
        return device_id.split(":")[0] == cls.get_implementation_id()

    @classmethod
    def get_device(cls, device_id: str) -> Device | None:
        """Get the device for the given device_id."""
        if not cls.is_responsible(device_id):
            return None

        devices = cls.search_for_devices()
        for d in devices:
            if d.device_id == device_id:
                return d
        return None

    @property
    @abc.abstractmethod
    def uid(self) -> str:
        """Return an identifier within the implementation context."""

    @staticmethod
    @abc.abstractmethod
    def get_device_type() -> str:
        """Return the implemented device type."""

    @property
    @abc.abstractmethod
    def manufacturer(self) -> str:
        """Return the device manufacturer."""

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """Return the product name."""

    @classmethod
    @abc.abstractmethod
    def get_implementation_id(cls) -> str:
        """Return an identifier for this implementation."""

    @classmethod
    @abc.abstractmethod
    def search_for_devices(cls: type[T]) -> list[T]:
        """Search for supported devices."""

    @abc.abstractmethod
    def summary(self) -> str:
        """Return a summary of the device and current status."""

    def __repr__(self) -> str:
        return f"{self.get_device_type()}: {self.device_id}"



"""Dummy implementation of PowerStrip."""
from __future__ import annotations

import random

from pyegps.powerstrip import PowerStrip


class FakePowerStrip(PowerStrip):
    """Dummy PowerStrip Device."""

    DEVICES: list[FakePowerStrip] = []

    def __init__(self, devId: str, number_of_sockets: int):
        """Initiate new DummyDevice."""
        self._devId = devId
        self._numberOfSockets = number_of_sockets
        self._status = [random.randint(0, 1) for _ in range(number_of_sockets)]

    # PowerStrip implementations
    @property
    def numberOfSockets(self):
        return self._numberOfSockets

    def switch_on(self, socket: int) -> None:
        super().switch_on(socket)
        self._status[socket] = 1

    def switch_off(self, socket: int) -> None:
        super().switch_off(socket)
        self._status[socket] = 0

    def get_status(self, socket: int) -> bool:
        super().get_status(socket)
        return self._status[socket] == 1

    # Device implementations
    @property
    def uid(self) -> str:
        """Return an identifier among DummyPowerStrip devices."""
        return self._devId

    @property
    def manufacturer(self) -> str:
        """Return the device manufacturer."""
        return "DummyDevices"

    @property
    def name(self) -> str:
        """Return the product name."""
        return "DummyPowerStrip"

    @classmethod
    def get_implementation_id(cls) -> str:
        """Return an identifier for this implementation."""
        return "DYPS"

    @classmethod
    def search_for_devices(cls) -> list[FakePowerStrip]:
        """Return the dummy devices, create new if necessary."""
        if len(cls.DEVICES) == 0:
            cls.DEVICES += [
                cls(devId, sockets)
                for devId, sockets in [("AA:BB:CC", 4), ("00:11:22", 2)]
            ]

        return cls.DEVICES



"""Controlling Energenie Power Strips."""
from __future__ import annotations
from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("pyegps")
except PackageNotFoundError:
    from ._version import __version__  # noqa


from .device import Device
from .usb.eg_powerstrip import PowerStripUSB
from .fakes.powerstrip import FakePowerStrip


DEVICE_IMPLEMENTATIONS: list[type[Device]] = [PowerStripUSB]


def search_for_devices(device_type: str | None = None) -> list[Device]:
    """Search and return all supported devices."""
    return [
        dev
        for impl in DEVICE_IMPLEMENTATIONS
        if device_type is None or impl.get_device_type() == device_type
        for dev in impl.search_for_devices()
    ]


def get_device_types() -> list[str]:
    """Return all supported device types."""
    return [impl.get_device_type() for impl in DEVICE_IMPLEMENTATIONS]


def get_device(device_id: str) -> Device | None:
    """Try to find the device for the given device_id and return it."""
    for impl in DEVICE_IMPLEMENTATIONS:
        dev_or_none = impl.get_device(device_id)
        if dev_or_none is not None:
            return dev_or_none
    return None


def use_fake_devices() -> None:
    """Let dummy devices appear in searches for devices."""
    DEVICE_IMPLEMENTATIONS.clear()
    DEVICE_IMPLEMENTATIONS.append(FakePowerStrip)



"""Simple command line interface for pyEGPS."""
from __future__ import annotations

import argparse

from . import search_for_devices, get_device_types, __version__
from .device import Device
from .powerstrip import PowerStrip
from .exceptions import EgpsException, InvalidSocketNumber


def print_status(dev: Device):
    """Print summary of given device."""
    print(dev.summary())


def list_devices(device_type: str = "all") -> int:
    """Print __repr__() for all devices."""
    for dev in search_for_devices():
        if device_type == "all" or dev.get_device_type() == device_type:
            print(dev)
    return 0


def cli(argList: list[str]) -> int:
    """Command Line Interface (CLI)."""
    parser = argparse.ArgumentParser(
        description="Simple command line interface for controlling EnerGenie Power Strips"
    )

    subparsers = parser.add_subparsers(dest="command")
    parser.add_argument(
        "--device",
        metavar="DEVICE_ID",
        nargs="?",
        help="Specify device[s] to use. Can be omitted if only one device is present.",
    )
    parser.add_argument("--debug", action="store_true", help="Print debug information.")
    parser.add_argument(
        "--version", action="store_true", help="Print debug information."
    )

    # List devices
    list_cmd = subparsers.add_parser("list", help="Search for supported devices.")

    # If several device types are defined, allow filtering by device type
    device_types = get_device_types()
    if len(device_types) > 1:
        list_cmd.add_argument(
            "--type", type=str, choices=["all"] + device_types, default="all"
        )

    set_cmd = subparsers.add_parser("set", help="Set status of sockets.")
    set_cmd.add_argument(
        "--on",
        type=int,
        nargs="*",
        action="append",
        metavar="SOCKET_NR",
        default=[],
        help="Turn SOCKET_NR on.",
    )
    set_cmd.add_argument(
        "--off",
        type=int,
        nargs="*",
        action="append",
        metavar="SOCKET_NR",
        default=[],
        help="Turn SOCKET_NR off.",
    )

    status_cmd = subparsers.add_parser("status", help="Get current device status.")
    status_cmd.add_argument(
        "SOCKET_NR",
        nargs="*",
        type=int,
        help="Status of SOCKET_NR, print summary if no SOCKET_NR is given.",
    )

    args = parser.parse_args(argList)

    if args.version:
        print(__version__)
        return 0

    if args.command == "list":
        dev_type = args.type if len(device_types) > 1 else "all"
        try:
            return list_devices(dev_type)
        except EgpsException as e:
            print(f"Error: {e}")
            return 1
    try:
        devices = search_for_devices(device_type="PowerStrip")
    except EgpsException as e:
        print(f"Error: {e}")
        return 1

    # filter devices if requested
    if args.device is not None:
        devices = [d for d in devices if d.device_id == args.device]

    if len(devices) == 0:
        print(
            "Couldn't find any devices{}.".format(
                f" with ids in {args.device}" if args.device is not None else "!"
            )
        )
        return 1

    if args.command == "set":
        on_sockets = [item for sublist in args.on for item in sublist]
        off_sockets = [item for sublist in args.off for item in sublist]
        if len(on_sockets) + len(off_sockets) == 0:
            print("Please specify at least one --on or --off argument")
            return 1
        for device in devices:
            assert isinstance(device, PowerStrip)
            for socket in on_sockets:
                try:
                    device.switch_on(socket)
                except InvalidSocketNumber:
                    print(
                        f"Device {device} has no socket {socket}. Known sockets: {list(range(device.numberOfSockets))}."
                    )
                    return 1
            for socket in off_sockets:
                try:
                    device.switch_off(socket)
                except InvalidSocketNumber:
                    print(
                        f"Device {device} has no socket {socket}. Known sockets: {list(range(device.numberOfSockets))}."
                    )
                    return 1
        return 0

    elif args.command == "status":
        for device in devices:
            assert isinstance(device, PowerStrip)
            if len(args.SOCKET_NR) == 0:
                print_status(device)
            else:
                status = []
                for socket in args.SOCKET_NR:
                    try:
                        status.append(device.get_status(socket))
                    except InvalidSocketNumber:
                        print(
                            f"Device {device} has no socket {socket}. Known sockets: {list(range(device.numberOfSockets))}."
                        )
                        return 1
                print(" ".join(map(lambda s: "on" if s else "off", status)))
        return 0
    else:
        parser.print_usage()
        return 1

    return 1



"""Test main functionality."""

from pyegps import DEVICE_IMPLEMENTATIONS


def test_implementations_have_unique_ids():
    """Test that all device implementations have their unique implementation id."""
    unique_ids = []
    for impl in DEVICE_IMPLEMENTATIONS:
        assert (impl_id := impl.get_implementation_id()) is not unique_ids
        unique_ids.append(impl_id)



"""Main pyEGPS tests."""
from __future__ import annotations

from array import array
import pytest
from unittest.mock import patch
import usb.core


from pyegps.usb.eg_powerstrip import PowerStripUSB
from pyegps.exceptions import MaximumConnectionTriesReached, MissingLibrary


def test_main(fakeUsbDevice):
    """Dummy Test."""
    # check that all device impl have a unique id
    ps = PowerStripUSB(fakeUsbDevice)
    assert ps.manufacturer == "AllFake"

    with pytest.raises(MaximumConnectionTriesReached):
        _ = ps.device_id

    with patch.object(
        fakeUsbDevice, "ctrl_transfer", return_value=array("B", [1, 2, 3])
    ):
        assert ps.device_id == ":".join(
            [ps.get_implementation_id()] + [format(x, "02x") for x in [1, 2, 3]]
        )


def test_main_error_check(fakeUsbDevice):
    """Dummy Test."""
    # check that all device impl have a unique id
    ps = PowerStripUSB(fakeUsbDevice)

    with patch.object(
        fakeUsbDevice,
        "ctrl_transfer",
        return_value=array("B", [1, 2, 3]),
        side_effect=usb.core.USBError("Error"),
    ):
        with pytest.raises(MaximumConnectionTriesReached):
            _ = ps.device_id

    with patch("usb.core.find", side_effect=usb.core.NoBackendError("Error")):
        with pytest.raises(MissingLibrary):
            PowerStripUSB.search_for_devices()



"""Command line interface (cli) tests for pyEGPS."""

import pytest
from unittest.mock import patch

from pyegps import cli
from pyegps.fakes.powerstrip import FakePowerStrip

fake_devices = FakePowerStrip.search_for_devices()


@pytest.mark.parametrize(
    ("device_id", "cli_args", "return_code"),
    [
        # Device not found
        ("XX:XX:XX:XX", ("set", "--on", "0"), 1),
        # No socket specified
        ("DYPS:00:11:22", ("set", "--on"), 1),
        # Device has sockets 0..3, set socket 3
        ("DYPS:AA:BB:CC", ("set", "--on", "3"), 0),
        # Device has sockets 0..1, set socket 3
        ("DYPS:00:11:22", ("set", "--on", "3"), 1),
        # Device has sockets 0..3, receive status 0..3
        ("DYPS:AA:BB:CC", ("status", "0", "1", "2", "3"), 0),
        # Device has sockets 0..1, receive status 0..3
        ("DYPS:00:11:22", ("status", "0", "1", "2", "3"), 1),
        # Device has sockets 0..3, receive status summary
        ("DYPS:AA:BB:CC", ("status",), 0),
    ],
)
def test_exit_codes(device_id: str, cli_args: tuple["str"], return_code: int) -> None:
    """Checking error handling via exit codes."""
    with patch("pyegps.cli.search_for_devices", return_value=fake_devices):
        assert cli.cli(("--device", device_id) + cli_args) == return_code


def test_outputs(capsys):
    """Checking consistency of setting and reading socket status."""
    with patch("pyegps.cli.search_for_devices", return_value=fake_devices):
        # set status and read if it the same
        cli.cli(
            ["--device", "DYPS:AA:BB:CC", "set", "--on", "0", "2", "--off", "1", "3"]
        )
        cli.cli(["--device", "DYPS:AA:BB:CC", "status", "0", "1", "2", "3"])
        captured = capsys.readouterr()
        assert captured.out.strip() == "on off on off"



"""Custom Exceptions of pyEGPS."""


class EgpsException(Exception):
    """General EGPS exception."""


class MissingLibrary(EgpsException):
    """System is a missing a library."""


class UsbError(EgpsException):
    """Can't access the usb device."""


class UnsupportedProducId(UsbError):
    """Device has an unsupported product id."""


class MaximumConnectionTriesReached(UsbError):
    """Couldn't connect to device."""


class InvalidSocketNumber(EgpsException):
    """Device has no such socket."""



"""Abstract PowerStrip Class."""
from __future__ import annotations
import abc
import logging

from .device import Device
from .exceptions import (
    InvalidSocketNumber,
)

_logger = logging.getLogger(__name__)


class PowerStrip(Device, abc.ABC):
    """Abstract class of a PowerStrip device."""

    @staticmethod
    def get_device_type() -> str:
        return "PowerStrip"

    @property
    @abc.abstractmethod
    def numberOfSockets(self) -> int:
        """Return number of controllable sockets."""
        return 0

    def get_status(self, socket: int) -> bool | None:
        """
        Get the status of the socket given by 'socket'.

        @param socket: socket number
        @return: status
        """
        if socket < 0 or socket > self.numberOfSockets - 1:
            raise InvalidSocketNumber
        return None

    def switch_off(self, socket: int) -> None:
        """
        Switches the socket with the given id off.

        @param socket: socket number
        """
        if socket < 0 or socket > self.numberOfSockets - 1:
            raise InvalidSocketNumber

        _logger.info(f"Socket {socket} switched off.")

    def switch_on(self, socket: int) -> None:
        """
        Switches the socket with the given id on.

        @param socket: socket number
        """
        if socket < 0 or socket > self.numberOfSockets - 1:
            raise InvalidSocketNumber

        _logger.info(f"Socket {socket} switched on.")

    def summary(self) -> str:
        s = self.__repr__() + "\n"
        for socket in range(self.numberOfSockets):
            s += f"Socket #{socket}: {'on' if self.get_status(socket) else 'off'}\n"
        return s



"""Implementation of USB PowerStrips by Energenie."""
from __future__ import annotations
import logging

from array import array
import usb.core
from usb.core import Device as UsbDevice, NoBackendError
from usb.util import (
    CTRL_IN,
    CTRL_OUT,
    CTRL_TYPE_CLASS,
    CTRL_RECIPIENT_INTERFACE,
    dispose_resources,
)

from ..device import Device
from ..powerstrip import PowerStrip

from ..exceptions import (
    MissingLibrary,
    MaximumConnectionTriesReached,
    UnsupportedProducId,
)

_logger = logging.getLogger(__name__)

USB_VENDOR_ID = 0x04B4
USB_PRODUCT_IDS = [0xFD10, 0xFD11, 0xFD12, 0xFD13, 0xFD15]
PRODUCT_SOCKET_RANGES = [(0, 0), (1, 1), (1, 4), (1, 4), (1, 4)]

USB_CTRL_TRANSFER_TRIES = 5
USB_CTRL_TRANSFER_TIMEOUT = 500

USB_HID_GET_REPORT = 0x01
USB_HID_SET_REPORT = 0x09
USB_HID_REPORT_INPUT = 0x01
USB_HID_REPORT_OUTPUT = 0x02
USB_HID_REPORT_FEATURE = 0x03


class PowerStripUSB(PowerStrip):
    """Represents an Energenie Power-Strip."""

    opened_devices: dict[str, Device] = {}
    port_dev_mapping: dict[str, str] = {}

    def __init__(self, dev: UsbDevice) -> None:
        """Initiate PowerStripUSB device.

        :param dev: usb device instance
        :type dev: usb.core.Device
        :raises UNSUPPORTED_PRODUCT_ID: Given usb device has unsupported product id.
        """
        self.productId: int = dev.idProduct

        if self.productId not in USB_PRODUCT_IDS:
            raise UnsupportedProducId

        self._dev: UsbDevice = dev
        self._uid: str | None = None

        _logger.debug(f"active kernel driver: {str(dev.is_kernel_driver_active(0))}")
        _logger.debug(f"address: {dev.address} bus: {dev.bus} ")

        minAddr, maxAddr = PRODUCT_SOCKET_RANGES[USB_PRODUCT_IDS.index(self.productId)]
        self._addrMapping: range = range(minAddr, maxAddr + 1)

    def release(self) -> None:
        """Release the usb device."""
        PowerStripUSB.opened_devices.pop(self.device_id, None)
        PowerStripUSB.port_dev_mapping.pop(f"{self._dev.bus}_{self._dev.address}", None)
        try:
            dispose_resources(self._dev)
        except Exception:
            pass

    @classmethod
    def get_implementation_id(cls) -> str:
        """Return an identifier for this implementation."""
        return "EGPS-USB"

    @property
    def uid(self) -> str:
        """Return an identifier for PowerStripUSB devices, read from firmware."""
        if self._uid is None:
            self._uid = self._read_device_id()
        return self._uid

    @property
    def manufacturer(self) -> str:
        """Return the manufacturer as read from the device."""
        return self._dev.manufacturer

    @property
    def name(self) -> str:
        """Return the product name as read from the device."""
        return self._dev.product

    @property
    def numberOfSockets(self) -> int:
        """Return number of controllable sockets."""
        return len(self._addrMapping)

    def get_status(self, socket: int) -> bool:
        """
        Get the status of the socket given by 'socket'.

        @param socket: socket number
        @return: status
        """
        super().get_status(socket)

        addr = self._addrMapping[socket]
        buf: bytes = self._get_feature_report(3 * addr)
        return (1 & buf[1]) == 1

    def switch_off(self, socket: int) -> None:
        """
        Switch the socket with the given id off.

        @param socket: socket number
        """
        super().switch_off(socket)

        addr = self._addrMapping[socket]
        buf = bytes([3 * addr, 0x00])
        self._set_feature_report(3 * addr, buf)

    def switch_on(self, socket: int) -> None:
        """
        Switch the socket with the given id on.

        @param socket: socket number
        """
        super().switch_on(socket)

        addr = self._addrMapping[socket]
        buf = bytes([3 * addr, 0x03])
        self._set_feature_report(3 * addr, buf)

    def _read_device_id(self) -> str:
        report_id: int = 0x01
        id = self._get_feature_report(report_id)
        _logger.debug(f"The device id is: {id!r}")
        return ":".join([format(x, "02x") for x in id])

    def _get_feature_report(self, report_id: int) -> bytes:
        bmRequestType: int = CTRL_IN | CTRL_TYPE_CLASS | CTRL_RECIPIENT_INTERFACE
        bRequest: int = USB_HID_GET_REPORT
        wValue: int = (USB_HID_REPORT_FEATURE << 8) | (report_id & 255)
        wIndex: int = 0
        wLength: int = 5
        ret = self._retry_ctrl_transfer(
            bmRequestType, bRequest, wValue, wIndex, wLength
        )
        assert isinstance(ret, bytes)
        return ret

    def _set_feature_report(self, report_id: int, data: bytes) -> int:
        bmRequestType: int = CTRL_OUT | CTRL_TYPE_CLASS | CTRL_RECIPIENT_INTERFACE
        bRequest: int = USB_HID_SET_REPORT
        wValue: int = (USB_HID_REPORT_FEATURE << 8) | (report_id & 255)
        wIndex: int = 0
        ret = self._retry_ctrl_transfer(bmRequestType, bRequest, wValue, wIndex, data)
        assert isinstance(ret, int)
        return ret

    def _retry_ctrl_transfer(
        self,
        bmRequestType: int,
        bRequest: int,
        wValue: int,
        wIndex: int,
        data_or_wLength: bytes | int,
    ) -> int | bytes:
        """Perform ctrl transfer and retry if not successful.

        The parameters bmRequestType, bRequest, wValue and wIndex are the same
        of the USB Standard Control Request format.

        Return the number of bytes written (for OUT transfers) or the data
        read (for IN transfers), as an array.array object.
        """
        _logger.debug(
            f"ctrl_transfer: {bmRequestType}, {bRequest}, {wValue}, {data_or_wLength!r}"
        )
        req_in: bool = (bmRequestType & CTRL_IN) == CTRL_IN
        for i in range(USB_CTRL_TRANSFER_TRIES):
            try:
                buf_or_len: bytes | int = self._dev.ctrl_transfer(
                    bmRequestType,
                    bRequest,
                    wValue,
                    wIndex,
                    data_or_wLength,
                    USB_CTRL_TRANSFER_TIMEOUT,
                )
            except usb.core.USBError as e:
                _logger.debug(f"ctrl_transfer: try number {i}, usb error: {e}")
                continue
            if req_in:
                assert isinstance(buf_or_len, array)
                if len(buf_or_len) == 0:
                    continue
                return bytes(buf_or_len)

            return int(buf_or_len)

        raise MaximumConnectionTriesReached

    @classmethod
    def search_for_devices(cls) -> list[PowerStripUSB]:
        """List the usb devices which have a known EG-PM product_id."""
        devices = []
        for prodId in USB_PRODUCT_IDS:
            try:
                devices += [
                    cls(dev)
                    for dev in usb.core.find(
                        find_all=True, idVendor=USB_VENDOR_ID, idProduct=prodId
                    )
                    if f"{dev.bus}_{dev.address}" not in cls.port_dev_mapping
                ]
            except NoBackendError:
                raise MissingLibrary
        return devices

    @classmethod
    def get_device(cls, device_id: str) -> Device | None:
        """Try to locate a specific EG-PM device.

        :param deviceId: The device specific firmware id.
        :type deviceId: str
        :return: _description_
        :rtype: PowerStripUSB | None
        """
        if device_id in cls.opened_devices:
            return cls.opened_devices[device_id]
        candidates = cls.search_for_devices()
        for dev in candidates:
            if dev.device_id == device_id:
                cls.opened_devices[device_id] = dev
                cls.port_dev_mapping[f"{dev._dev.bus}_{dev._dev.address}"] = device_id
                return dev
        return None
